@@ -11,7 +11,7 @@ from sqlalchemy.orm import mapper, query
 from sqlalchemy.orm.interfaces import *
 from sqlalchemy.orm import session as sessionlib
 from sqlalchemy.orm import util as mapperutil
-import sets, random
+import random
 
 
 class ColumnLoader(LoaderStrategy):
@@ -20,13 +20,13 @@ class ColumnLoader(LoaderStrategy):
         self.columns = self.parent_property.columns
         self._should_log_debug = logging.is_debug_enabled(self.logger)
         
-    def setup_query(self, context, eagertable=None, **kwargs):
+    def setup_query(self, context, eagertable=None, parentclauses=None, **kwargs):
         for c in self.columns:
-            if eagertable is not None:
-                context.statement.append_column(eagertable.corresponding_column(c))
+            if parentclauses is not None:
+                context.statement.append_column(parentclauses.aliased_column(c))
             else:
                 context.statement.append_column(c)
-
+        
     def init_class_attribute(self):
         self.logger.info("register managed attribute %s on class %s" % (self.key, self.parent.class_.__name__))
         coltype = self.columns[0].type
@@ -41,8 +41,12 @@ class ColumnLoader(LoaderStrategy):
 ColumnLoader.logger = logging.class_logger(ColumnLoader)
 
 class DeferredColumnLoader(LoaderStrategy):
-    """describes an object attribute that corresponds to a table column, which also
-    will "lazy load" its value from the table.  this is per-column lazy loading."""
+    """Describes an object attribute that corresponds to a table
+    column, which also will *lazy load* its value from the table.
+
+    This is per-column lazy loading.
+    """
+    
     def init(self):
         super(DeferredColumnLoader, self).init()
         self.columns = self.parent_property.columns
@@ -73,6 +77,10 @@ class DeferredColumnLoader(LoaderStrategy):
         def lazyload():
             if self._should_log_debug:
                 self.logger.debug("deferred load %s group %s" % (mapperutil.attribute_str(instance, self.key), str(self.group)))
+
+            if not mapper.has_identity(instance):
+                return None
+
             try:
                 pk = self.parent.pks_by_table[self.columns[0].table]
             except KeyError:
@@ -116,6 +124,7 @@ class DeferredOption(StrategizedOption):
     def __init__(self, key, defer=False):
         super(DeferredOption, self).__init__(key)
         self.defer = defer
+
     def get_strategy_class(self):
         if self.defer:
             return DeferredColumnLoader
@@ -125,20 +134,8 @@ class DeferredOption(StrategizedOption):
 class AbstractRelationLoader(LoaderStrategy):
     def init(self):
         super(AbstractRelationLoader, self).init()
-        self.primaryjoin = self.parent_property.primaryjoin
-        self.secondaryjoin = self.parent_property.secondaryjoin
-        self.secondary = self.parent_property.secondary
-        self.foreignkey = self.parent_property.foreignkey
-        self.mapper = self.parent_property.mapper
-        self.select_mapper = self.mapper.get_select_mapper()
-        self.target = self.parent_property.target
-        self.select_table = self.parent_property.mapper.select_table
-        self.loads_polymorphic = self.target is not self.select_table
-        self.uselist = self.parent_property.uselist
-        self.cascade = self.parent_property.cascade
-        self.attributeext = self.parent_property.attributeext
-        self.order_by = self.parent_property.order_by
-        self.remote_side = self.parent_property.remote_side
+        for attr in ['primaryjoin', 'secondaryjoin', 'secondary', 'foreign_keys', 'mapper', 'select_mapper', 'target', 'select_table', 'loads_polymorphic', 'uselist', 'cascade', 'attributeext', 'order_by', 'remote_side', 'polymorphic_primaryjoin', 'polymorphic_secondaryjoin', 'direction']:
+            setattr(self, attr, getattr(self.parent_property, attr))
         self._should_log_debug = logging.is_debug_enabled(self.logger)
         
     def _init_instance_attribute(self, instance, callable_=None):
@@ -151,6 +148,7 @@ class AbstractRelationLoader(LoaderStrategy):
 class NoLoader(AbstractRelationLoader):
     def init_class_attribute(self):
         self._register_attribute(self.parent.class_)
+
     def process_row(self, selectcontext, instance, row, identitykey, isnew):
         if isnew:
             if not self.is_default or len(selectcontext.options):
@@ -163,7 +161,8 @@ NoLoader.logger = logging.class_logger(NoLoader)
 class LazyLoader(AbstractRelationLoader):
     def init(self):
         super(LazyLoader, self).init()
-        (self.lazywhere, self.lazybinds, self.lazyreverse) = self._create_lazy_clause(self.parent.unjoined_table, self.primaryjoin, self.secondaryjoin, self.foreignkey, self.remote_side, self.mapper.select_table)
+        (self.lazywhere, self.lazybinds, self.lazyreverse) = self._create_lazy_clause(self)
+
         # determine if our "lazywhere" clause is the same as the mapper's
         # get() clause.  then we can just use mapper.get()
         self.use_get = not self.uselist and query.Query(self.mapper)._get_clause.compare(self.lazywhere)
@@ -210,7 +209,7 @@ class LazyLoader(AbstractRelationLoader):
             # to possibly save a DB round trip
             if self.use_get:
                 ident = []
-                for primary_key in self.mapper.pks_by_table[self.mapper.mapped_table]:
+                for primary_key in self.select_mapper.pks_by_table[self.select_mapper.mapped_table]:
                     bind = self.lazyreverse[primary_key]
                     ident.append(params[bind.key])
                 return session.query(self.mapper).get(ident)
@@ -247,15 +246,18 @@ class LazyLoader(AbstractRelationLoader):
                 # to load data into it.
                 sessionlib.attribute_manager.reset_instance_attribute(instance, self.key)
 
-    def _create_lazy_clause(self, table, primaryjoin, secondaryjoin, foreignkey, remote_side, select_table):
+    def _create_lazy_clause(cls, prop, reverse_direction=False):
+        (primaryjoin, secondaryjoin, remote_side) = (prop.polymorphic_primaryjoin, prop.polymorphic_secondaryjoin, prop.remote_side)
+        
         binds = {}
         reverse = {}
-        def column_in_table(table, column):
-            return table.corresponding_column(column, raiseerr=False, keys_ok=False) is not None
 
-        if remote_side is None or len(remote_side) == 0:
-            remote_side = foreignkey
-            
+        def should_bind(targetcol, othercol):
+            if reverse_direction:
+                return targetcol in remote_side
+            else:
+                return othercol in remote_side
+
         def find_column_in_expr(expr):
             if not isinstance(expr, sql.ColumnElement):
                 return None
@@ -263,75 +265,78 @@ class LazyLoader(AbstractRelationLoader):
             class FindColumnInColumnClause(sql.ClauseVisitor):
                 def visit_column(self, c):
                     columns.append(c)
-            expr.accept_visitor(FindColumnInColumnClause())
+            FindColumnInColumnClause().traverse(expr)
             return len(columns) and columns[0] or None
         
-        def col_in_collection(column, collection):
-            for c in collection:
-                if column.shares_lineage(c):
-                    return True
-            else:
-                return False
-                
         def bind_label():
+            # TODO: make this generation deterministic
             return "lazy_" + hex(random.randint(0, 65535))[2:]
+
         def visit_binary(binary):
             leftcol = find_column_in_expr(binary.left)
             rightcol = find_column_in_expr(binary.right)
             if leftcol is None or rightcol is None:
                 return
-            circular = leftcol.table is rightcol.table
-            if ((not circular and column_in_table(table, leftcol)) or (circular and col_in_collection(rightcol, remote_side))):
+            if should_bind(leftcol, rightcol):
                 col = leftcol
                 binary.left = binds.setdefault(leftcol,
-                        sql.bindparam(bind_label(), None, shortname=leftcol.name, type=binary.right.type))
+                        sql.bindparam(bind_label(), None, shortname=leftcol.name, type=binary.right.type, unique=True))
                 reverse[rightcol] = binds[col]
 
-            if (leftcol is not rightcol) and ((not circular and column_in_table(table, rightcol)) or (circular and col_in_collection(leftcol, remote_side))):
+            # the "left is not right" compare is to handle part of a join clause that is "table.c.col1==table.c.col1",
+            # which can happen in rare cases (test/orm/relationships.py RelationTest2)
+            if leftcol is not rightcol and should_bind(rightcol, leftcol):
                 col = rightcol
                 binary.right = binds.setdefault(rightcol,
-                        sql.bindparam(bind_label(), None, shortname=rightcol.name, type=binary.left.type))
+                        sql.bindparam(bind_label(), None, shortname=rightcol.name, type=binary.left.type, unique=True))
                 reverse[leftcol] = binds[col]
 
         lazywhere = primaryjoin.copy_container()
         li = mapperutil.BinaryVisitor(visit_binary)
-        lazywhere.accept_visitor(li)
+        li.traverse(lazywhere)
         
         if secondaryjoin is not None:
             secondaryjoin = secondaryjoin.copy_container()
-            secondaryjoin.accept_visitor(sql_util.ClauseAdapter(select_table))
             lazywhere = sql.and_(lazywhere, secondaryjoin)
-        else:
-            lazywhere.accept_visitor(sql_util.ClauseAdapter(select_table))
-            
-        LazyLoader.logger.info("create_lazy_clause " + str(lazywhere))
+ 
+        if hasattr(cls, 'parent_property'):
+            LazyLoader.logger.info(str(cls.parent_property) + " lazy loading clause " + str(lazywhere))
         return (lazywhere, binds, reverse)
-
+    _create_lazy_clause = classmethod(_create_lazy_clause)
+    
 LazyLoader.logger = logging.class_logger(LazyLoader)
 
 
-        
 class EagerLoader(AbstractRelationLoader):
-    """loads related objects inline with a parent query."""
+    """Loads related objects inline with a parent query."""
+    
     def init(self):
         super(EagerLoader, self).init()
-        if self.parent.isa(self.select_mapper):
-            raise exceptions.ArgumentError("Error creating eager relationship '%s' on parent class '%s' to child class '%s': Cant use eager loading on a self referential relationship." % (self.key, repr(self.parent.class_), repr(self.mapper.class_)))
-        self.parent._eager_loaders.add(self.parent_property)
+        if self.parent.isa(self.mapper):
+            raise exceptions.ArgumentError(
+                "Error creating eager relationship '%s' on parent class '%s' "
+                "to child class '%s': Cant use eager loading on a self "
+                "referential relationship." %
+                (self.key, repr(self.parent.class_), repr(self.mapper.class_)))
+        if self.is_default:
+            self.parent._eager_loaders.add(self.parent_property)
 
         self.clauses = {}
         self.clauses_by_lead_mapper = {}
 
     class AliasedClauses(object):
-        """defines a set of join conditions and table aliases which are aliased on a randomly-generated
-        alias name, corresponding to the connection of an optional parent AliasedClauses object and a 
-        target mapper.
+        """Defines a set of join conditions and table aliases which
+        are aliased on a randomly-generated alias name, corresponding
+        to the connection of an optional parent AliasedClauses object
+        and a target mapper.
         
-        EagerLoader has a distinct AliasedClauses object per parent AliasedClauses object,
-        so that all paths from one mapper to another across a chain of eagerloaders generates a distinct
-        chain of joins.  The AliasedClauses objects are generated and cached on an as-needed basis.
+        EagerLoader has a distinct AliasedClauses object per parent
+        AliasedClauses object, so that all paths from one mapper to
+        another across a chain of eagerloaders generates a distinct
+        chain of joins.  The AliasedClauses objects are generated and
+        cached on an as-needed basis.
         
-        e.g.:
+        E.g.::
         
             mapper A -->
                 (EagerLoader 'items') --> 
@@ -339,7 +344,7 @@ class EagerLoader(AbstractRelationLoader):
                         (EagerLoader 'keywords') --> 
                             mapper C
             
-            will generate:
+        will generate::
             
             EagerLoader 'items' --> {
                 None : AliasedClauses(items, None, alias_suffix='AB34')        # mappera JOIN mapperb_AB34
@@ -351,32 +356,31 @@ class EagerLoader(AbstractRelationLoader):
                         AliasedClauses(keywords, items, alias_suffix='8F44')   # mapperb_AB34 JOIN mapperc_8F44
             ]
         """
+        
         def __init__(self, eagerloader, parentclauses=None):
+            self.id = (parentclauses is not None and (parentclauses.id + "/") or '') + str(eagerloader.parent_property)
             self.parent = eagerloader
             self.target = eagerloader.select_table
-            self.eagertarget = eagerloader.select_table.alias()
+            self.eagertarget = eagerloader.select_table.alias(self._aliashash("/target"))
+            self.extra_cols = {}
             
             if eagerloader.secondary:
-                self.eagersecondary = eagerloader.secondary.alias()
+                self.eagersecondary = eagerloader.secondary.alias(self._aliashash("/secondary"))
                 self.aliasizer = sql_util.Aliasizer(eagerloader.target, eagerloader.secondary, aliases={
                         eagerloader.target:self.eagertarget,
                         eagerloader.secondary:self.eagersecondary
                         })
-                self.eagersecondaryjoin = eagerloader.secondaryjoin.copy_container()
-                if eagerloader.loads_polymorphic:
-                    self.eagersecondaryjoin.accept_visitor(sql_util.ClauseAdapter(eagerloader.select_table))
-                self.eagersecondaryjoin.accept_visitor(self.aliasizer)
-                self.eagerprimary = eagerloader.primaryjoin.copy_container()
-                self.eagerprimary.accept_visitor(self.aliasizer)
+                self.eagersecondaryjoin = eagerloader.polymorphic_secondaryjoin.copy_container()
+                self.aliasizer.traverse(self.eagersecondaryjoin)
+                self.eagerprimary = eagerloader.polymorphic_primaryjoin.copy_container()
+                self.aliasizer.traverse(self.eagerprimary)
             else:
-                self.eagerprimary = eagerloader.primaryjoin.copy_container()
-                if eagerloader.loads_polymorphic:
-                    self.eagerprimary.accept_visitor(sql_util.ClauseAdapter(eagerloader.select_table))
+                self.eagerprimary = eagerloader.polymorphic_primaryjoin.copy_container()
                 self.aliasizer = sql_util.Aliasizer(self.target, aliases={self.target:self.eagertarget})
-                self.eagerprimary.accept_visitor(self.aliasizer)
+                self.aliasizer.traverse(self.eagerprimary)
 
             if parentclauses is not None:
-                self.eagerprimary.accept_visitor(parentclauses.aliasizer)
+                parentclauses.aliasizer.traverse(self.eagerprimary)
 
             if eagerloader.order_by:
                 self.eager_order_by = self._aliasize_orderby(eagerloader.order_by)
@@ -384,18 +388,38 @@ class EagerLoader(AbstractRelationLoader):
                 self.eager_order_by = None
 
             self._row_decorator = self._create_decorator_row()
+        
+        def aliased_column(self, column):
+            """return the aliased version of the given column, creating a new label for it if not already
+            present in this AliasedClauses eagertable."""
 
+            conv = self.eagertarget.corresponding_column(column, raiseerr=False)
+            if conv:
+                return conv
+
+            if column in self.extra_cols:
+                return self.extra_cols[column]
+            
+            aliased_column = column.copy_container()
+            sql_util.ClauseAdapter(self.eagertarget).traverse(aliased_column)
+            alias = self._aliashash(column.name)
+            aliased_column = aliased_column.label(alias)
+            self._row_decorator.map[column] = alias
+            self.extra_cols[column] = aliased_column
+            return aliased_column
+            
+        def _aliashash(self, extra):
+            """return a deterministic 4 digit hash value for this AliasedClause's id + extra."""
+            # use the first 4 digits of an MD5 hash
+            return "anon_" + util.hash(self.id + extra)[0:4]
+            
         def _aliasize_orderby(self, orderby, copy=True):
             if copy:
-                orderby = [o.copy_container() for o in util.to_list(orderby)]
+                return self.aliasizer.copy_and_process(util.to_list(orderby))
             else:
                 orderby = util.to_list(orderby)
-            for i in range(0, len(orderby)):
-                if isinstance(orderby[i], schema.Column):
-                    orderby[i] = self.eagertarget.corresponding_column(orderby[i])
-                else:
-                    orderby[i].accept_visitor(self.aliasizer)
-            return orderby
+                self.aliasizer.process_list(orderby)
+                return orderby
 
         def _create_decorator_row(self):
             class EagerRowAdapter(object):
@@ -415,6 +439,7 @@ class EagerLoader(AbstractRelationLoader):
                 map[parent] = c
                 map[parent._label] = c
                 map[parent.name] = c
+            EagerRowAdapter.map = map
             return EagerRowAdapter
 
         def _decorate_row(self, row):
@@ -427,7 +452,8 @@ class EagerLoader(AbstractRelationLoader):
         self.parent_property._get_strategy(LazyLoader).init_class_attribute()
         
     def setup_query(self, context, eagertable=None, parentclauses=None, parentmapper=None, **kwargs):
-        """add a left outer join to the statement thats being constructed"""
+        """Add a left outer join to the statement thats being constructed."""
+        
         if parentmapper is None:
             localparent = context.mapper
         else:
@@ -464,6 +490,8 @@ class EagerLoader(AbstractRelationLoader):
         except KeyError:
             clauses = EagerLoader.AliasedClauses(self, parentclauses)
             self.clauses[parentclauses] = clauses
+            
+        if context.mapper not in self.clauses_by_lead_mapper:
             self.clauses_by_lead_mapper[context.mapper] = clauses
 
         if self.secondaryjoin is not None:
@@ -477,17 +505,19 @@ class EagerLoader(AbstractRelationLoader):
 
         if clauses.eager_order_by:
             statement.order_by(*util.to_list(clauses.eager_order_by))
-        elif getattr(statement, 'order_by_clause', None):
-            clauses._aliasize_orderby(statement.order_by_clause, False)
                 
         statement.append_from(statement._outerjoin)
         for value in self.select_mapper.props.values():
             value.setup(context, eagertable=clauses.eagertarget, parentclauses=clauses, parentmapper=self.select_mapper)
 
     def _create_row_processor(self, selectcontext, row):
-        """create a 'row processing' function that will apply eager aliasing to the row.
+        """Create a *row processing* function that will apply eager
+        aliasing to the row.
+        
+        Also check that an identity key can be retrieved from the row,
+        else return None.
+        """
         
-        also check that an identity key can be retrieved from the row, else return None."""
         # check for a user-defined decorator in the SelectContext (which was set up by the contains_eager() option)
         if selectcontext.attributes.has_key((EagerLoader, self.parent_property)):
             # custom row decoration function, placed in the selectcontext by the 
@@ -500,7 +530,7 @@ class EagerLoader(AbstractRelationLoader):
                 # decorate the row according to the stored AliasedClauses for this eager load
                 clauses = self.clauses_by_lead_mapper[selectcontext.mapper]
                 decorator = clauses._row_decorator
-            except KeyError:
+            except KeyError, k:
                 # no stored AliasedClauses: eager loading was not set up in the query and
                 # AliasedClauses never got initialized
                 return None
@@ -511,13 +541,19 @@ class EagerLoader(AbstractRelationLoader):
             identity_key = self.mapper.identity_key_from_row(decorated_row)
             # and its good
             return decorator
-        except KeyError:
+        except KeyError, k:
             # no identity key - dont return a row processor, will cause a degrade to lazy
+            if self._should_log_debug:
+                self.logger.debug("could not locate identity key from row '%s'; missing column '%s'" % (repr(decorated_row), str(k)))
             return None
 
     def process_row(self, selectcontext, instance, row, identitykey, isnew):
-        """receive a row.  tell our mapper to look for a new object instance in the row, and attach
-        it to a list on the parent instance."""
+        """Receive a row.
+
+        Tell our mapper to look for a new object instance in the row,
+        and attach it to a list on the parent instance.
+        """
+        
         if self in selectcontext.recursion_stack:
             return
         
@@ -578,6 +614,7 @@ class EagerLazyOption(StrategizedOption):
     def __init__(self, key, lazy=True):
         super(EagerLazyOption, self).__init__(key)
         self.lazy = lazy
+
     def process_query_property(self, context, prop):
         if self.lazy:
             if prop in context.eager_loaders:
@@ -585,6 +622,7 @@ class EagerLazyOption(StrategizedOption):
         else:
             context.eager_loaders.add(prop)
         super(EagerLazyOption, self).process_query_property(context, prop)
+
     def get_strategy_class(self):
         if self.lazy:
             return LazyLoader
@@ -592,14 +630,27 @@ class EagerLazyOption(StrategizedOption):
             return EagerLoader
         elif self.lazy is None:
             return NoLoader
+
 EagerLazyOption.logger = logging.class_logger(EagerLazyOption)
 
 class RowDecorateOption(PropertyOption):
-    def __init__(self, key, decorator=None):
+    def __init__(self, key, decorator=None, alias=None):
         super(RowDecorateOption, self).__init__(key)
         self.decorator = decorator
+        self.alias = alias
+
     def process_selection_property(self, context, property):
+        if self.alias is not None and self.decorator is None:
+            if isinstance(self.alias, basestring):
+                self.alias = property.target.alias(self.alias)
+            def decorate(row):
+                d = {}
+                for c in property.target.columns:
+                    d[c] = row[self.alias.corresponding_column(c)]
+                return d
+            self.decorator = decorate
         context.attributes[(EagerLoader, property)] = self.decorator
+
 RowDecorateOption.logger = logging.class_logger(RowDecorateOption)
         
 
